"""
Created on Aug 29, 2014

@author: tmahrt
"""

import os
from os.path import join

import wave
import audioop

from pyjuliusalign import utils

# The default framerate in Julius
DEFAULT_FRAMERATE = 16000


class IncompatibleSampleFrequencyError(Exception):
    def __init__(self, actualSamplingRate, targetSamplingRate):
        super(IncompatibleSampleFrequencyError, self).__init__()
        self.actualSR = actualSamplingRate
        self.targetSR = targetSamplingRate

    def __str__(self):
        return (
            "File's sampling rate=%s but system requires=%s\n "
            "Please resample audio files or ensure that sox is in "
            "your path" % (self.actualSR, self.targetSR)
        )


def getSoundFileDuration(fn):
    """
    Returns the duration of a wav file (in seconds)
    """
    audiofile = wave.open(fn, "r")

    params = audiofile.getparams()
    framerate = params[2]
    nframes = params[3]

    duration = float(nframes) / framerate
    return duration


def extractSubwav(fn, outputFN, startT, endT, singleChannelFlag, soxPath=None):

    if soxPath is None:
        soxPath = "sox"  # Assumes it is in the user's path

    path, fnNoPath = os.path.split(fn)
    resampledAudioPath = join(path, "resampledAudio")
    resampledFN = join(resampledAudioPath, fnNoPath)

    audiofile = wave.open(fn, "r")
    params = audiofile.getparams()
    nchannels = params[0]
    sampwidth = params[1]
    framerate = params[2]
    comptype = params[4]
    compname = params[5]

    # If you are not using the default Julius training model, you might
    # need to change the value here.  Results will fail if the sampling
    # frequency is different.
    if framerate != DEFAULT_FRAMERATE:
        if not os.path.exists(resampledFN):
            utils.makeDir(resampledAudioPath)
            sr = str(DEFAULT_FRAMERATE)
            soxCmd = "%s %s -r %s %s rate -v 96k" % (soxPath, fn, sr, resampledFN)
            os.system(soxCmd)

        if not os.path.exists(resampledFN):
            raise IncompatibleSampleFrequencyError(framerate, DEFAULT_FRAMERATE)

        audiofile = wave.open(resampledFN, "r")
        params = audiofile.getparams()
        nchannels = params[0]
        sampwidth = params[1]
        framerate = params[2]
        comptype = params[4]
        compname = params[5]

    # Extract the audio frames
    audiofile.setpos(int(framerate * startT))
    audioFrames = audiofile.readframes(int(framerate * (endT - startT)))

    # Convert to single channel if needed
    if singleChannelFlag is True and nchannels > 1:
        audioFrames = audioop.tomono(audioFrames, sampwidth, 1, 0)
        nchannels = 1

    outParams = [nchannels, sampwidth, framerate, len(audioFrames), comptype, compname]

    outWave = wave.open(outputFN, "w")
    outWave.setparams(outParams)
    outWave.writeframes(audioFrames)


def splitStereoAudio(path, fn, outputPath=None):

    if outputPath is None:
        outputPath = join(path, "split_audio")

    if not os.path.exists(outputPath):
        os.mkdir(outputPath)

    name = os.path.splitext(fn)[0]

    fnFullPath = join(path, fn)
    leftOutputFN = join(outputPath, "%s_L.wav" % name)
    rightOutputFN = join(outputPath, "%s_R.wav" % name)

    audiofile = wave.open(fnFullPath, "r")

    params = audiofile.getparams()
    sampwidth = params[1]
    nframes = params[3]
    audioFrames = audiofile.readframes(nframes)

    for leftFactor, rightFactor, outputFN in (
        (1, 0, leftOutputFN),
        (0, 1, rightOutputFN),
    ):

        monoAudioFrames = audioop.tomono(
            audioFrames, sampwidth, leftFactor, rightFactor
        )
        params = tuple(
            [
                1,
            ]
            + list(params[1:])
        )

        outputAudiofile = wave.open(outputFN, "w")
        outputAudiofile.setparams(params)
        outputAudiofile.writeframes(monoAudioFrames)



# -*- coding: utf-8 -*-
"""
Created on Aug 20, 2014

@author: tmahrt
"""

import os
from os.path import join
import datetime
import subprocess

import Levenshtein

from pyjuliusalign import jProcessingSnippet
from pyjuliusalign import audioScripts
from pyjuliusalign import utils
from pyjuliusalign import convertKana
from pyjuliusalign import yomi2voca

PHONE = "phones"
WORD = "words"
UTTERANCE = "utterances"


class JuliusRunError(Exception):
    def __init__(self, juliusPath):
        super(JuliusRunError, self).__init__()
        self.path = juliusPath

    def __str__(self):
        return "Tried to run julius at location but it does not exist: \n%s" % self.path


class JuliusAlignmentError(Exception):
    def __str__(self):
        return "Failed to align: "


class JuliusScriptExecutionFailed(Exception):
    def __init__(self, cmdList):
        super(JuliusScriptExecutionFailed, self).__init__()
        self.cmdList = cmdList

    def __str__(self):
        errorStr = (
            "\ Execution Failed.  Please check the following:\n"
            "- Perl and the julius4.pl script exist in the "
            "locations specified\n"
            "- you have edited the 'user configuration' part of "
            "'segment_julius4.pl'\n"
            "- script arguments are correct\n\n"
            "If you can't locate the problem, I recommend using "
            "absolute paths rather than relative "
            "paths and using paths without spaces in any folder "
            "or file names\n\n"
            "Here is the command that python attempted to run:\n"
        )
        cmdTxt = " ".join(self.cmdList)
        return errorStr + cmdTxt


def runJuliusAlignment(resourcePath, juliusScriptPath, perlPath, loggerFd):

    if not os.path.exists(juliusScriptPath):
        raise JuliusRunError(juliusScriptPath)

    resourcePath = os.path.abspath(resourcePath)

    cmdList = [perlPath, juliusScriptPath, resourcePath]
    print(cmdList)
    myProcess = subprocess.Popen(cmdList, stdout=loggerFd)

    if myProcess.wait():
        raise JuliusScriptExecutionFailed(cmdList)


def parseJuliusOutput(juliusOutputFn):
    with open(juliusOutputFn, "r") as fd:
        juliusOutput = fd.read()

    returnList = []
    for row in juliusOutput.split("\n"):
        row = row.strip()
        if row == "":
            continue
        row = row.split(" ")
        start = float(row[0])
        stop = float(row[1])
        label = row[2]
        returnList.append((start, stop, label))

    return returnList


def mapJuliusPronunciationToCabocha(juliusPhonesTxt, cabochaPhonesByWord):
    """
    Aligns the phones in a julius pronunciation list into phones chunked into words by cabocha

    The phonetisation is a little different, so the mapping tries to do so gracefully.

    Basically, modify cabochaPhonesByWord to have the same number of phones as juliusPhonesTxt.
    This is done by finding the edits necessary to make the two strings the same and then applying
    those edits.
    Once the two strings contain the same number of phones, dump the phones from julius into the slots for
    the phones inside of the modified cabocha words.
    """

    def _buildWordIndicies(cabochaPhonesByWord):
        startI = 0
        wordIndicies = []
        for word in cabochaPhonesByWord:
            wordIndicies.append([startI, startI + len(word)])
            startI += len(word)

        return wordIndicies

    def _getWordForCharIndex(indiciesForWords, targetI):
        returnI = None
        for i, indicies in enumerate(indiciesForWords):
            start, stop = indicies
            if targetI >= start and targetI < stop:
                returnI = i
                break
        return returnI

    cabochaPhonesByWord = [phones.replace(":", "") for phones in cabochaPhonesByWord]
    # TODO: 'w' is treated as a phone but its not--should this logic live elsewhere?
    cabochaPhonesByWord = [phones.replace("w o", "o") for phones in cabochaPhonesByWord]
    cabochaPhonesTxt = " ".join(cabochaPhonesByWord)

    # Mutate cabochaPhonesByWord to contain the same number
    # of phones as juliusPhonesTxt
    edits = Levenshtein.editops(cabochaPhonesTxt, juliusPhonesTxt)
    wordIndicies = _buildWordIndicies(cabochaPhonesByWord)
    for operation, startIndex, _ in edits:
        wordI = _getWordForCharIndex(wordIndicies, startIndex)
        if operation == "delete":
            cabochaPhonesByWord[wordI] = cabochaPhonesByWord[wordI][:-1]
        elif operation == "insert":
            cabochaPhonesByWord[wordI] += "-"

    # Chunk juliusPhonesByWord according to the number
    # of phones in the now aligned cabochaPhonesByWord
    juliusPhonesByWord = []
    startI = 0
    for wordNum, phones in enumerate(cabochaPhonesByWord):
        endI = startI + len(phones)
        juliusWordPhones = juliusPhonesTxt[startI:endI]

        juliusPhonesByWord.append(juliusWordPhones)
        startI = endI + 1  # Add 1 space for the space between words

    return juliusPhonesByWord


def juliusAlignCabocha(
    dataList,
    wavPath,
    wavFN,
    juliusScriptPath,
    soxPath,
    perlPath,
    silenceFlag,
    forceEndTimeFlag,
    forceMonophoneAlignFlag,
):
    """
    Given utterance-level timing and a wav file, phone-align the audio

    dataList is the formatted output of cabocha of the form
    [startTime, endTime, wordList, kanaList, romajiList]
    """
    tmpOutputPath = join(wavPath, "align_tmp")
    utils.makeDir(tmpOutputPath)

    logFn = join(tmpOutputPath, "align_log_" + str(datetime.datetime.now()) + ".txt")
    loggerFd = open(logFn, "w")

    utils.makeDir(tmpOutputPath)

    tmpTxtFN = join(tmpOutputPath, "tmp.txt")
    tmpWavFN = join(tmpOutputPath, "tmp.wav")
    tmpOutputFN = join(tmpOutputPath, "tmp.lab")

    entryDict = {}
    for aspect in [UTTERANCE, WORD, PHONE]:
        entryDict[aspect] = []

    # one speech interval at a time
    numTotalPhones = 0
    numPhonesFailedToAlign = 0
    numIntervals = 0
    numFailedIntervals = 0

    # intervalStart, intervalEnd, line, wordList, kanaList, romajiList
    for rowTuple in dataList:
        intervalStart = rowTuple[0]
        intervalEnd = rowTuple[1]
        line = rowTuple[2]
        wordList = rowTuple[3]
        kanaList = convertKana.convertKataToKana(
            [char for row in rowTuple[4] for char in row]
        )
        cabochaPhonesByWord = rowTuple[5]

        if line.strip() != "":
            entryDict[UTTERANCE].append((str(intervalStart), str(intervalEnd), line))

        if len([word for word in wordList if word != ""]) == 0:
            continue

        assert intervalStart < intervalEnd

        # Create romajiTxt (for forced alignment) and
        # phoneList (for the textgrid)
        # Phones broken up by word
        tmpRomajiList = []
        tmpFlattenedRomajiList = []
        juliusPhones = yomi2voca.convert(kanaList)
        juilusPronunciationByWord = mapJuliusPronunciationToCabocha(
            juliusPhones, cabochaPhonesByWord
        )
        for row in juilusPronunciationByWord:
            rowList = row.split(" ")
            tmpRomajiList.append(rowList)
            tmpFlattenedRomajiList.extend(rowList)

        numWords = len(wordList)
        wordTimeList = [[] for i in range(numWords)]

        # No forced-alignment if there is no romaji
        romajiTxt = "".join(kanaList)
        if romajiTxt.strip() == "":
            continue

        # Encapsulate each phone string in boundary silence
        #    - in my limited experience, this messes up the output even more
        if silenceFlag:
            romajiTxt = "silB " + romajiTxt + " silE"

        # Save temporary transcript and wav files for interval
        with open(tmpTxtFN, "w") as fd:
            fd.write(romajiTxt)

        audioScripts.extractSubwav(
            join(wavPath, wavFN),
            tmpWavFN,
            intervalStart,
            intervalEnd,
            singleChannelFlag=False,
            soxPath=soxPath,
        )

        # Run forced alignment
        runJuliusAlignment(tmpOutputPath, juliusScriptPath, perlPath, loggerFd)

        # Get the output (timestamps for each phone)
        numIntervals += 1
        try:
            matchList = parseJuliusOutput(tmpOutputFN)
        except JuliusAlignmentError:
            if forceMonophoneAlignFlag is True and numWords == 1:
                # One phone occupies the whole interval
                matchList = [(0.0, (intervalEnd - intervalStart) * 100)]
            else:
                numPhonesFailedToAlign += numWords
                numFailedIntervals += 1
                print(
                    "Failed to align: %s - %f - %f"
                    % ("".join(romajiList), intervalStart, intervalEnd)
                )
                continue

        adjustedPhonList = [
            [intervalStart + start, intervalStart + stop, label]
            for start, stop, label in matchList
        ]

        # Julius is conservative in estimating the final vowel.  Stretch it
        # to be the length of the utterance
        if forceEndTimeFlag:
            adjustedPhonList[-1][1] = intervalEnd

        entryDict[PHONE].extend(adjustedPhonList)

        # Get the bounding indicies for the phones in each word
        phoneToWordIndexList = []
        phonesSoFar = 0
        for i in range(len(wordList)):
            numPhones = len(tmpRomajiList[i])
            phoneToWordIndexList.append((phonesSoFar, phonesSoFar + numPhones - 1))
            phonesSoFar += numPhones

        # If julius uses a silence model and we don't, then adjust our timings
        phoneListFromJulius = [label for _, _, label in adjustedPhonList]
        if "silB" in phoneListFromJulius and "silB" not in tmpFlattenedRomajiList:
            phoneToWordIndexList = [
                (startI + 1, endI + 1) for startI, endI in phoneToWordIndexList
            ]
            lastI = phoneToWordIndexList[-1][1]
            phoneToWordIndexList = (
                [(0, 0)] + phoneToWordIndexList + [(lastI + 1, lastI + 1)]
            )
            wordList = [""] + wordList + [""]

        # Store the words
        print(tmpRomajiList)
        print(cabochaPhonesByWord)
        for i in range(len(wordList)):
            startI, stopI = phoneToWordIndexList[i]
            try:
                entryDict[WORD].append(
                    (
                        adjustedPhonList[startI][0],
                        adjustedPhonList[stopI][1],
                        wordList[i],
                    )
                )
            except IndexError:
                raise

        numTotalPhones += numWords

    statList = [
        numPhonesFailedToAlign,
        numTotalPhones,
        numFailedIntervals,
        numIntervals,
    ]
    return entryDict, statList


def formatTextForJulius(line, cabochaEncoding, cabochaPath):
    """Prepares a single line of text, processed by cabocha, for use in julius"""
    unidentifiedUtterance = 0
    unnamedEntity = 0

    origLine = line

    # Clean up the line before it gets processed
    # Not sure what "・" is but cabocha doesn't like it
    for char in [
        u"（",
        u"）",
        u" ",
        u"．",
        u"？",
        u"「",
        u"」",
        u"［",
        u"］",
        u"＠Ｗ",
        u"＠Ｓ",
        u"＜",
        u"＞",
        u" ",
        u"。",
    ]:
        line = line.replace(char, "")

    # Used to split names?
    for char in [u"・", u"·"]:
        line = line.replace(char, " ")

    line = line.strip()

    try:
        (tmpWordList, tmpKanaList, tmpRomajiList) = jProcessingSnippet.getChunkedKana(
            line, cabochaEncoding, cabochaPath
        )
    except (jProcessingSnippet.ChunkingError, jProcessingSnippet.NonKatakanaError) as e:
        print(u"%s, %s" % (str(e), origLine))
        unidentifiedUtterance = 1
    except jProcessingSnippet.UnidentifiedJapaneseText as e:
        # Maybe specific to my corpus?
        if all([char == u"X" for char in e.word]):
            unnamedEntity = 1
        else:
            print(u"%s" % str(e))
            unidentifiedUtterance = 1
    except jProcessingSnippet.EmptyStrError as e:
        pass
    except Exception:
        print(line)
        raise
    line = line.replace(u",", u"")

    return (
        line,
        ",".join(tmpWordList),
        ",".join(tmpKanaList),
        ",".join(tmpRomajiList),
        unidentifiedUtterance,
        unnamedEntity,
        len(tmpWordList),
    )



"""
Created on Aug 29, 2014

@author: tmahrt
"""

import os
import functools


def _getMatchFunc(pattern):
    """
    An unsophisticated pattern matching function
    """

    # '#' Marks word boundaries, so if there is more than one we need to do
    #    something special to make sure we're not mis-representings them
    assert pattern.count("#") < 2

    def startsWith(subStr, fullStr):
        return fullStr[: len(subStr)] == subStr

    def endsWith(subStr, fullStr):
        return fullStr[-1 * len(subStr) :] == subStr

    def inStr(subStr, fullStr):
        return subStr in fullStr

    # Selection of the correct function
    if pattern[0] == "#":
        pattern = pattern[1:]
        cmpFunc = startsWith

    elif pattern[-1] == "#":
        pattern = pattern[:-1]
        cmpFunc = endsWith

    else:
        cmpFunc = inStr

    return functools.partial(cmpFunc, pattern)


def findFiles(
    path,
    filterPaths=False,
    filterExt=None,
    filterPattern=None,
    skipIfNameInList=None,
    stripExt=False,
):

    fnList = os.listdir(path)

    if filterPaths is True:
        fnList = [
            folderName
            for folderName in fnList
            if os.path.isdir(os.path.join(path, folderName))
        ]

    if filterExt is not None:
        splitFNList = [
            [
                fn,
            ]
            + list(os.path.splitext(fn))
            for fn in fnList
        ]
        fnList = [fn for fn, name, ext in splitFNList if ext == filterExt]

    if filterPattern is not None:
        splitFNList = [
            [
                fn,
            ]
            + list(os.path.splitext(fn))
            for fn in fnList
        ]
        matchFunc = _getMatchFunc(filterPattern)
        fnList = [fn for fn, name, ext in splitFNList if matchFunc(name)]

    if skipIfNameInList is not None:
        targetNameList = [os.path.splitext(fn)[0] for fn in skipIfNameInList]
        fnList = [fn for fn in fnList if os.path.splitext(fn)[0] not in targetNameList]

    if stripExt is True:
        fnList = [os.path.splitext(fn)[0] for fn in fnList]

    fnList.sort()
    return fnList


def makeDir(path):
    """Makes the desired directory if it doesn't already exist"""
    if not os.path.exists(path):
        os.mkdir(path)


def divide(numerator, denominator, zeroValue):
    if denominator == 0:
        retValue = zeroValue
    else:
        retValue = numerator / float(denominator)

    return retValue



# -*- coding: utf-8 -*-
"""
Created on Aug 20, 2014

@author: tmahrt

katakanaChart.txt, hiraganaChart.txt, formdamage(), parseChart(),
cabocha() and jReads()
were all taken with modification from the project
https://github.com/kevincobain2000/jProcessing
# Copyright (c) 2011, Pulkit Kathuria
# All rights reserved.

Modifications were made specifically for prepping data in use with julius

All other code is from me
"""

import os
import subprocess
import tempfile

import xml.etree.cElementTree as etree

from pyjuliusalign import convertKana


class UnidentifiedJapaneseText(Exception):
    def __init__(self, sentence, word):
        super(UnidentifiedJapaneseText, self).__init__()
        self.sentence = sentence
        self.word = word

    def __str__(self):
        return u"No match in dictionary for word '%s' in sentence: \n'%s'" % (
            self.word,
            self.sentence,
        )


class ChunkingError(Exception):
    """Raised when a katakana string cannot be parsed correctly"""

    def __init__(self, txt):
        super(ChunkingError, self).__init__()
        self.textStr = txt

    def __str__(self):
        return u"Chunking error for string: \n %s" % self.textStr


class EmptyStrError(Exception):
    def __str__(self):
        return "Empty string passed in"


class NonKatakanaError(Exception):
    def __init__(self, char, utterance):
        super(NonKatakanaError, self).__init__()
        self.char = char
        self.utterance = utterance

    def __str__(self):
        return u"Wrongly interpreted character '%s' as kana in utterance:\n%s" % (
            unicode(self.char),
            unicode(self.utterance),
        )


class CabochaNotRunError(Exception):
    def __init__(self, cabochaCmd):
        super(CabochaNotRunError, self).__init__()
        self.cabochaCmd = cabochaCmd

    def __str__(self):
        return (
            u"Cabocha not installed or not in path.  "
            "Attempted to run command:\n%s" % self.cabochaCmd
        )


class CabochaOutputError(Exception):
    def __init__(self, outputStr):
        super(CabochaOutputError, self).__init__()
        self.outputStr = outputStr

    def __str__(self):
        return (
            "Unexpected error in cabocha output "
            "(possibly a problem with cabocha).  See error:\n%s" % self.outputStr
        )


class TextDecodingError(Exception):
    def __init__(self, encoding):
        super(TextDecodingError, self).__init__()
        self.encoding = encoding

    def __str__(self):
        print(self.encoding)
        return (
            "Cabocha tried to decode text with the encoding '%s' but it failed. "
            "If you specify a different encoding in "
            "alignFromTextgrid.convertCorpusToKanaAndRomaji() that will probably "
            "fix the problem." % self.encoding
        )


def _formdamage(sent):
    rectify = []
    for ch in sent:
        try:
            rectify.append(ch.encode("euc-jp"))
        except:
            pass
    return "".join(rectify)


def cabocha(sentence, cabochaEncoding, cabochaPath=None):
    """
    Sends the sentence to be processed by cabocha

    /cabochaEncoding/ is either 'jis-shift', 'utf-8', or 'euc-jp',
    whichever cabocha was installed with
    """

    if cabochaPath is None:
        cabochaPath = "cabocha"  # Assumes cabocha is in user path

    temp = tempfile.NamedTemporaryFile(delete=False)

    try:
        sentence = sentence.encode(cabochaEncoding)
    except:
        sentence = _formdamage(sentence)

    temp.write(sentence)
    temp.close()

    command = [cabochaPath, "-f", "3", temp.name]
    try:
        process = subprocess.Popen(command, stdout=subprocess.PIPE)
    except OSError:
        raise CabochaNotRunError(" ".join(command))
    output = process.communicate()[0]
    os.unlink(temp.name)

    try:
        retStr = str(output, cabochaEncoding)
    except UnicodeDecodeError:
        raise TextDecodingError(cabochaEncoding)

    return retStr


def jReads(target_sent, cabochaEncoding, cabochaPath):

    xmlStr = cabocha(target_sent, cabochaEncoding, cabochaPath).encode("utf-8")
    #     print(target_sent)
    #     print(xmlStr)
    try:
        sentence = etree.fromstring(xmlStr)
    except etree.ParseError:
        raise CabochaOutputError(xmlStr)

    retJReadsToks = []
    retWordList = []

    keyList = list(convertKana.kataToRomajiDict.keys())
    validKatakanaList = [
        u"ャ",
        u"ュ",
        u"ョ",
        u"ッ",
        u"ァ",
        u"ィ",
        u"ゥ",
        u"ェ",
        u"ォ",
        u"ー",
    ] + keyList
    #     validHiraganaList = [u"ゃ", u"ゅ", u"ょ", u"っ"] + convertKana.kanaToRomajiDict.keys()
    for chunk in sentence:
        jReadsToks = []
        wordList = []
        for tok in chunk.findall("tok"):

            # Determine if this word is punctuation (and if so, skip it)
            # Contains syntactic and morphological information
            featureStr = tok.get("feature")
            featureList = featureStr.split(",")
            if featureList[0] == u"記号":
                continue

            # Don't process empty words (can happen a lot,
            # depending on the annotation scheme)
            word = tok.text.strip()
            if word == "":
                continue

            kana = featureList[-1]

            if kana != "*":
                jReadsToks.append(kana)
            else:
                # Map all hiragana to katakana (in what situations would a
                # hiragana word not appear in our dictionary?
                word = convertKana.convertKanaToKata(word)

                # If the text is all katakana, keep it,
                # otherwise, assume an error
                if all([char in validKatakanaList for char in word]):
                    jReadsToks.append(word)
                else:
                    raise UnidentifiedJapaneseText(target_sent, word)

            wordList.append(word)

        retJReadsToks.append(jReadsToks)
        retWordList.append(wordList)

    return retJReadsToks, retWordList


def getChunkedKana(string, cabochaEncoding, cabochaPath):

    if string == "":
        raise EmptyStrError()

    kanaList, wordList = jReads(string, cabochaEncoding, cabochaPath)
    kanaList = ["".join(subList) for subList in kanaList if len(subList) > 0]
    wordList = ["".join(subList) for subList in wordList if len(subList) > 0]

    # Special case - sometimes 'ー' ends up in its own word or at the start
    # of a word, but really it just makes the last vowel longer.  If its
    # by itself, attach it to the previous word
    def mergeTailingVowel(tmpList):
        retList = []
        for word in tmpList:
            if word[0] == u"ー":
                retList[-1] += word
            else:
                retList.append(word)
        return retList

    kanaList = mergeTailingVowel(kanaList)
    wordList = mergeTailingVowel(wordList)

    vowelList = [u"ア", u"イ", u"ウ", u"エ", u"オ"]
    vowelModifierDict = {u"ァ": u"a", u"ィ": u"i", u"ゥ": u"u", u"ェ": u"e", u"ォ": u"o"}
    yModifierDict = {u"ャ": u"ya", u"ュ": u"yu", u"ョ": u"yo"}

    romanjiedTextList = []
    for word in kanaList:

        # Sanitize kana
        try:
            kanaInputList = chunkKatakana(word)
        except IndexError:
            raise ChunkingError(word)

        # Convert kana to romanji
        romanjiList = []
        for kana in kanaInputList:

            # Over-write the previous vowel (foreign words)
            if kana in vowelModifierDict.keys():
                romanjiList[-1] = romanjiList[-1][:-1] + vowelModifierDict[kana]

            # Last vowel is a long vowel
            elif kana == u"ー":  # Long vowel
                # e.g. 'ホーー' becomes "ho::" or 'ふうんー' becomes "hu:n:"
                #  -- both bad
                if romanjiList[-1][-1] != ":" and romanjiList[-1][-1] != "N":
                    romanjiList[-1] += ":"

                # Should get recognized as a single vowel by the forced aligner

            # Normal case
            else:
                # Single-phone characters
                if kana == u"ン" or kana in vowelList:
                    romanjiList.append(convertKana.kataToRomajiDict[kana])
                else:
                    if kana in yModifierDict.keys():  # e.g. 'ィ' in 'ティム'
                        romanjiList.pop(-1)
                        syllable = yModifierDict[kana]
                    else:
                        try:
                            # Normal case for two-phone characters
                            syllable = convertKana.kataToRomajiDict[kana]
                        except KeyError:
                            raise NonKatakanaError(kana, string)

                    # The consonant (e.g. 's' or 'sh')
                    romanjiList.append(syllable[:-1])
                    # The vowel
                    romanjiList.append(syllable[-1])

        romanjiedTextList.append(" ".join(romanjiList))

    return wordList, kanaList, romanjiedTextList


def chunkKatakana(kanaStr):
    """
    Chunks katakana for use in the katakana chart
    """
    kanaList = list(kanaStr)

    returnList = []
    i = 0
    while i < len(kanaStr):
        kana = kanaList[i]
        # Some palatalized consonants are in our chart
        # (notably 'jy' and 'shy' are missing)
        kanaFlag = kana in [u"ャ", u"ュ", u"ョ"]
        if kanaFlag and returnList[-1] + kana in convertKana.kataToRomajiDict.keys():
            returnList[-1] += kana
        elif kana == u"ッ":  # Geminate consonant
            pass
        else:  # Normal case
            returnList.append(kana)
        i += 1

    return returnList



import io
import os
from os.path import join

jNlp_KATAKANA_CHART_PATH = join(os.path.dirname(__file__), "katakanaChart.txt")
jNlp_HIRAGANA_CHART_PATH = join(os.path.dirname(__file__), "hiraganaChart.txt")


def parseChart(chartFN):
    """
    @return chartDict
    ��� ==> g,a
    ��� ==> k,i
    ������ ==> k,ya
    Similarily for Hiragana
    @setrofim : http://www.python-forum.
     org/pythonforum/viewtopic.php?f=3&t=31935
    """
    with io.open(chartFN, "r", encoding="utf-8") as fd:
        chart = fd.read()

    lines = chart.split("\n")
    chartDict = {}
    output = {}
    col_headings = lines.pop(0).split()
    for line in lines:
        cells = line.split()
        for i, c in enumerate(cells[1:]):
            output[c] = cells[0], col_headings[i]

    for k in sorted(output.keys()):
        # @k = katakana
        # @r = first romaji in row
        # @c = concatinating romaji in column
        r, c = output[k]
        if k == "X":
            continue
        romaji = "".join([item.replace("X", "") for item in [r, c]])
        chartDict[k] = romaji

    return chartDict


def _invertDict(tmpDict):
    """Flips key-value relationship in a dictionary"""
    retDict = {}
    for key, value in tmpDict.items():
        retDict[value] = key

    return retDict


def _getKataToKanaDict():
    retDict = {
        u"ャ": u"ゃ",
        u"ュ": u"ゅ",
        u"ョ": u"ょ",
        u"ッ": u"っ",
        u"ァ": u"ぁ",
        u"ィ": u"ぃ",
        u"ゥ": u"ぅ",
        u"ェ": u"ぇ",
        u"ォ": u"ぉ",
    }

    for kata in kataToRomajiDict.keys():
        kana = romajiToKanaDict[kataToRomajiDict[kata]]
        retDict[kata] = kana

    return retDict


kataToRomajiDict = parseChart(jNlp_KATAKANA_CHART_PATH)
kanaToRomajiDict = parseChart(jNlp_HIRAGANA_CHART_PATH)
romajiToKataDict = _invertDict(kataToRomajiDict)
romajiToKanaDict = _invertDict(kanaToRomajiDict)
kataToKanaDict = _getKataToKanaDict()
kanaToKataDict = _invertDict(kataToKanaDict)


def convertKanaToKata(inputStr):
    """Input hiragana and return the corresponding string in katakana"""
    retStr = u""
    for char in inputStr:
        if char in kanaToKataDict.keys():
            char = kanaToKataDict[char]
        retStr += char

    return retStr


def convertKataToKana(inputStr):
    """Input hiragana and return the corresponding string in katakana"""
    retStr = u""
    for char in inputStr:
        if char in kataToKanaDict.keys():
            char = kataToKanaDict[char]
        retStr += char

    return retStr



# -*- coding: utf-8 -*-
"""
Created on Aug 6, 2014

@author: tmahrt

Contains a basic workflow of functions that may be useful in force
alignment but are not strictly necessary.  If your input is from
textgrid files or files in the form ("start time", "stop time", "text")
you may find these functions useful.
"""

import os
from os.path import join
import io

from pyjuliusalign import utils
from pyjuliusalign import juliusAlignment
from pyjuliusalign import audioScripts

from praatio import tgio


def textgridToCSV(inputPath, outputPath, outputExt=".csv", tierName="utterances"):
    utils.makeDir(outputPath)

    for fn in utils.findFiles(inputPath, filterExt=".TextGrid"):
        tg = tgio.openTextgrid(join(inputPath, fn))
        tier = tg.tierDict[tierName]
        outputList = []
        for start, stop, label in tier.entryList:
            outputList.append("%s,%s,%s" % (start, stop, label))

        name = os.path.splitext(fn)[0]
        outputTxt = "\n".join(outputList)
        outputFN = join(outputPath, "%s%s" % (name, outputExt))
        with io.open(outputFN, "w", encoding="utf-8") as fd:
            fd.write(outputTxt)


def convertCorpusToKanaAndRomaji(
    inputPath, outputPath, cabochaEncoding, cabochaPath=None, encoding="cp932"
):
    """
    Reduces a corpus of typical Japanese text to both kana and romaji

    Each line of input should be of the form:
    startTime, stopTime, Japanese text
    """
    utils.makeDir(outputPath)

    numUnnamedEntities = 0
    numUnidentifiedUtterances = 0
    numWordsProcessedWithNoError = 0

    fnList = utils.findFiles(inputPath, filterExt=".txt")
    for fn in fnList:
        with io.open(join(inputPath, fn), "rU", encoding=encoding) as fd:
            text = fd.read()
        textList = text.split("\n")

        numUnnamedEntitiesForFN = 0
        numUnidentifiedUtterancesForFN = 0
        speakerInfoList = []
        for line in textList:
            line = line.strip()
            try:
                startTime, stopTime, line = line.split(",", 2)
            except ValueError:
                print("error")
                continue
            origLine = line

            dataPrepTuple = juliusAlignment.formatTextForJulius(
                line, cabochaEncoding, cabochaPath
            )

            (
                line,
                tmpWordList,
                tmpKanaList,
                tmpRomajiList,
                unidentifiedUtterance,
                unnamedEntity,
                tmpWordCount,
            ) = dataPrepTuple

            numUnnamedEntities += unnamedEntity
            numUnidentifiedUtterances += unidentifiedUtterance
            numWordsProcessedWithNoError += tmpWordCount

            name = os.path.splitext(fn)[0]
            outputList = [
                u"%s,%s,%s" % (name, startTime, stopTime),
                origLine,
                tmpWordList,
                tmpKanaList,
                tmpRomajiList,
            ]
            outputStr = ";".join(outputList)

            speakerInfoList.append(outputStr)

        if numUnnamedEntities > 0 or numUnidentifiedUtterances > 0:
            print(fn)
            print("Number of unnamed entities for fn: %d" % numUnnamedEntitiesForFN)
            print(
                "Number of unidentified utterances for fn: %d"
                % numUnidentifiedUtterancesForFN
            )

        numUnnamedEntities += numUnnamedEntitiesForFN
        numUnidentifiedUtterances += numUnidentifiedUtterancesForFN

        with io.open(join(outputPath, fn), "w", encoding="utf-8") as fd:
            fd.write("\n".join(speakerInfoList))

    print("\n")
    print("Number of transcripts converted: %d" % len(fnList))
    print("Number of unnamed entities: %d" % numUnnamedEntities)
    print("Number of unidentified utterances: %d" % numUnidentifiedUtterances)
    print("Number of words processed without error: %d" % numWordsProcessedWithNoError)


def forceAlignFile(
    speakerList,
    wavPath,
    wavNameDict,
    txtPath,
    txtFN,
    outputPath,
    outputWavName,
    juliusScriptPath,
    soxPath,
    perlPath,
):
    """

    Normally:
    speakerList = [name]
    and
    wavNameDict = {name:"name.wav"}

    But, if you have multiple speakers for each file (assuming audio is synced)
    e.g. in a stereo audio situation:
    speakerList=["L","R"]
    and
    wavNameDict={"L":"%s_%s.wav" % (name, "L"), "R":"%s_%s.wav" % (name, "R")}
    """

    utils.makeDir(outputPath)

    # Formatted output of cabocha
    with io.open(join(txtPath, txtFN), "r", encoding="utf-8") as fd:
        data = fd.read()
    dataList = data.split("\n")
    dataList = [
        [subRow.split(",") for subRow in row.split(";")]
        for row in dataList
        if row != ""
    ]

    dataDict = {speaker: [] for speaker in speakerList}

    # Undoing the unnecessary split that just happened
    for timingInfo, line, wordList, kanaList, romajiList in dataList:
        line = ",".join(line)

        speaker, startTimeStr, endTimeStr = timingInfo
        speaker, startTime, endTime = (
            speaker.strip(),
            float(startTimeStr),
            float(endTimeStr),
        )

        dataDict[speaker].append(
            [startTime, endTime, line, wordList, kanaList, romajiList]
        )

    # Do the forced alignment
    speakerEntryDict = {}
    numPhonesFailedAlignment = 0
    numPhones = 0
    numFailedIntervals = 0
    numIntervals = 0
    for speaker in speakerList:
        output = juliusAlignment.juliusAlignCabocha(
            dataDict[speaker],
            wavPath,
            wavNameDict[speaker],
            juliusScriptPath,
            soxPath,
            perlPath,
            False,
            True,
            True,
        )
        speakerEntryDict[speaker], statList = output

        numPhonesFailedAlignment += statList[0]
        numPhones += statList[1]
        numFailedIntervals += statList[2]
        numIntervals += statList[3]

    # All durations should be the same
    inputWavFN = next(iter(wavNameDict.values()))
    maxDuration = audioScripts.getSoundFileDuration(join(wavPath, inputWavFN))

    # Create tiers and textgrids from the output of the alignment
    tg = tgio.Textgrid()
    for speaker in speakerList:
        for aspect in [
            juliusAlignment.UTTERANCE,
            juliusAlignment.WORD,
            juliusAlignment.PHONE,
        ]:

            tierName = "%s_%s" % (aspect, speaker)
            tier = tgio.IntervalTier(
                tierName, speakerEntryDict[speaker][aspect], minT=0, maxT=maxDuration
            )
            tg.addTier(tier)

    tg.save(join(outputPath, outputWavName + ".TextGrid"))

    return (numPhonesFailedAlignment, numPhones, numFailedIntervals, numIntervals)


def forceAlignCorpus(
    wavPath, txtPath, outputPath, juliusScriptPath=None, soxPath=None, perlPath=None
):
    """Force aligns every file and prints out summary statistics"""
    totalNumPhonesFailed = 0
    totalNumPhones = 0

    totalNumIntervalsFailed = 0
    totalNumIntervals = 0

    utils.makeDir(outputPath)

    for name in utils.findFiles(txtPath, filterExt=".txt", stripExt=True):
        wavNameDict = {name: "%s.wav" % name}
        output = forceAlignFile(
            [
                name,
            ],
            wavPath,
            wavNameDict,
            txtPath,
            name + ".txt",
            outputPath,
            name,
            juliusScriptPath,
            soxPath,
            perlPath,
        )

        (numPhonesFailedAlignment, numPhones, numFailedIntervals, numIntervals) = output

        percentFailed = utils.divide(numPhonesFailedAlignment, numPhones, 0) * 100
        percentFailedIntervals = utils.divide(numFailedIntervals, numIntervals, 0) * 100
        print(
            "%d intervals of %d total intervals (%0.2f%%) and %d phones "
            "of %d total phones (%0.2f%%) successfully aligned for %s"
            % (
                numIntervals - numFailedIntervals,
                numIntervals,
                100 * (1 - percentFailedIntervals),
                numPhones - numPhonesFailedAlignment,
                numPhones,
                100 * (1 - percentFailed),
                name,
            )
        )

        totalNumPhonesFailed += numPhonesFailedAlignment
        totalNumPhones += numPhones

        totalNumIntervalsFailed += numFailedIntervals
        totalNumIntervals += numIntervals

    totalPercentFailed = utils.divide(totalNumPhonesFailed, totalNumPhones, 0) * 100
    totalPercentFailedIntervals = (
        utils.divide(totalNumIntervalsFailed, totalNumIntervals, 0) * 100
    )
    print("====Summary====")
    print(
        "%d intervals of %d total intervals (%0.2f%%) and %d phones of "
        "%d total phones (%0.2f%%) successfully aligned"
        % (
            totalNumIntervals - totalNumIntervalsFailed,
            totalNumIntervals,
            100 * (1 - totalPercentFailedIntervals),
            totalNumPhones - totalNumPhonesFailed,
            totalNumPhones,
            100 * (1 - totalPercentFailed),
        )
    )



# -*- coding: utf-8 -*-
"""
Created on Aug 19, 2014

@author: tmahrt
"""

from os.path import join
import cPickle
import io


class UnidicTool(object):
    """
    A Japanese pronunciation dictionary developed by Ninjal
    """

    def __init__(self, unidicPath):
        self.data = None
        self.cachedData = None

        self.unidicPath = join(unidicPath, "lex.csv")
        self.unidicCachePath = join(unidicPath, "lex.pickle")

    def lookup(self, word):
        if self.data is None:
            with io.open(self.unidicPath, "r", encoding="utf-8") as fd:
                self.data = fd.read()

        startI = 0
        matchList = []
        while True:
            try:
                matchI = self.data.index(u"\n" + word, startI) + 1
                endI = self.data.index(u"\n", matchI)
            except ValueError:
                break

            matchList.append(self.data[matchI + len(word) + 1 : endI].split(","))
            startI = matchI

        return matchList

    def saveCache(self):
        cPickle.dump(self.cachedData, open(self.unidicCachePath, "w"))



#!/usr/bin/env python
# encoding: utf-8
"""
Created on Aug 29, 2014

@author: tmahrt
"""
from setuptools import setup
import io

setup(
    name="pyjuliusalign",
    version="3.0.2",
    author="Tim Mahrt",
    author_email="timmahrt@gmail.com",
    url="https://github.com/timmahrt/pyJuliusAlign",
    package_dir={"pyjuliusalign": "pyjuliusalign"},
    packages=["pyjuliusalign"],
    package_data={"pyjuliusalign": ["hiraganaChart.txt", "katakanaChart.txt"]},
    license="LICENSE",
    description="A helper library for doing forced-alignment in Japanese with Julius.",
    install_requires=["praatio ~= 4.1", "python-Levenshtein"],
    long_description=io.open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
)
